
pub const INTENSITY_DESCRIPTOR_NAMES: [&str; 9] = [
    "intensity_max",
    "intensity_min",
    "intensity_integrated",
    "intensity_mean",
    "intensity_std",
    "intensity_median",
    "intensity_mad",
    "intensity_skewness",
    "intensity_kurtosis",
];

pub const MOMENTS_DESCRIPTOR_NAMES: [&str; 24] = [
    "moments_raw_m00", 
    "moments_raw_m10",
    "moments_raw_m01",
    "moments_raw_m11",
    "moments_raw_m20",
    "moments_raw_m02",
    "moments_raw_m21",
    "moments_raw_m12",
    "moments_raw_m30",
    "moments_raw_m03",
    "moments_central_u11",
    "moments_central_u20",
    "moments_central_u02",
    "moments_central_u21",
    "moments_central_u12",
    "moments_central_u30",
    "moments_central_u03",
    "moments_hu_i1",
    "moments_hu_i2",
    "moments_hu_i3",
    "moments_hu_i4",
    "moments_hu_i5",
    "moments_hu_i6",
    "moments_hu_i7"
];

pub const POLYGON_DESCRIPTOR_NAMES: [&str; 23] = [
    "centroid_x",
    "centroid_y",
    "area",
    "area_bbox",
    "area_convex",
    "elongation",
    "perimeter",
    "thread_length",
    "thread_width",
    "solidity",
    "extent",
    "form_factor",
    "eccentricity",
    "major_axis_length",
    "minor_axis_length",
    "curl_major",
    "curl_bbox",
    "equivalent_diameter",
    "minimum_radius",
    "maximum_radius",
    "mean_radius",
    "feret_diameter_maximum",
    "feret_diameter_minimum",
];

pub const TEXTURE_DESCRIPTOR_NAMES: [&str; 13] = [
    "texture_energy",
    "texture_contrast",
    "texture_correlation",
    "texture_sum_of_squares",
    "texture_inverse_difference_moment",
    "texture_sum_average",
    "texture_sum_variance",
    "texture_sum_entropy",
    "texture_entropy",
    "texture_difference_variance",
    "texture_difference_entropy",
    "texture_information_measure_of_correlation_1",
    "texture_information_measure_of_correlation_2",
];

pub const ZERNIKE_DESCRIPTOR_NAMES: [&str; 30] = [
    "zernike_00",
    "zernike_11",
    "zernike_20",
    "zernike_22",
    "zernike_31",
    "zernike_33",
    "zernike_40",
    "zernike_42",
    "zernike_44",
    "zernike_51",
    "zernike_53",
    "zernike_55",
    "zernike_60",
    "zernike_62",
    "zernike_64",
    "zernike_66",
    "zernike_71",
    "zernike_73",
    "zernike_75",
    "zernike_77",
    "zernike_80",
    "zernike_82",
    "zernike_84",
    "zernike_86",
    "zernike_88",
    "zernike_91",
    "zernike_93",
    "zernike_95",
    "zernike_97",
    "zernike_99",
];
